/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
TODO:
The letters of the non-standard comm. commands are a bit too random,
fix it before it is too late!
*/


/*
  This file deserves a little explanation!

  Instead of creating a huge circular buffer for uart data,
  a smaller circular buffer 'command_queue' are used instead.
  This buffer contains the address and arguments to functions which
  can fetch bytes from any place in the memory.

  eg. MegaTune sends an 'A' to request realtime variables:
  1) Uart interrupt calls comm()
  2) comm pushes the address of sendRTvar and the number of
     bytes to send (22) onto the command queue and enables tx-interrupt.
  3) tx-interrupt reads the command queue, executes the function and
     decrements the number of bytes to send (and stores it in the
     command queue again). If number equals zero remove the function
     from the queue and start over. If the queue is empty, then disable
     interrupt.
*/

   /*!
   * \addtogroup g_serial_comm Serial Communication Routines
   */
/*@{*/

#include "global.h"
#include "comm.h"
#include "vems.h"
#include "aim.h"
#include "ringbuf.h"
#include "storage.h"
#include "actuators.h"
#include "ve.h"
#include "iac.h"
#include "adc.h"
#include "lcd.h"
#include "ego.h"
#include "fuelcalc.h"
#include "helpers.h"
#include "benchmark.h"
#include "wbo2.h"
#include "boostcontrol.h"
#include "camcontrol.h"
#include "slcan.h"
#include "kb.h"
#include "spi.h"
#include "firmware_revision.h"
#include "knock.h"
#include "fuelcontrol.h"
#include "my_make.h"

#include "triggerframe.h"

#ifdef MAINLOOP_TASKER
#include "mainloop.h"
#endif

#ifdef A340E
#include "a340e.h"
#endif
#ifdef GRAY_DECODER
#include "gray_decoder.h"
#endif

#ifdef CAN_MCP2515
#include "mss6x_idle_etc.h"
#endif

#include <avr/io.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "varstr.h"
#include "varstr_ext.h"
#include "varstr_shared.h"
extern const char f_version[] PROGMEM;

#include "arithm.c"
#include <util/crc16.h>

#if MODBUS_ON_COMM == 2
#include "injector.h"
#include "modbus_rtu_slave.h"
#endif

struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];
uint8_t cmd_queue_tail, cmd_queue_head;

struct com_t com;

const char vems_string[] PROGMEM = "VEMS Genboard v3, ";

/* baudrate change pending flag. Ignored when 0,
 * otherwise it contains the new baudrate divisor */
uint8_t new_baudrate;

// TABLES_TO_DUMP1 lives in menu.c because otherwise gcc fucks up flashsize by +500 bytes beyond imagination
const uint8_t TABLES_TO_DUMP2[] PROGMEM = {
'n',  2+IGNADV_SIZE_KPA-1

#ifdef HWMAP
,'h', HWMAP_LINES-1
#endif

,'i', 2+ MAP_CORR_SIZE_KPA-1
,'w', 2+ LAMBDA_SIZE_KPA-1
#ifdef CLT_IGN_ADJUST
,'s', 2+3-1
#else
,'s', 2-1
#endif
,'p', 3-1 // 3 anytrim1_array lines (17 bytes)
};

 /** send echo via UART0,
 *allow bytes coming from UART1 within next 2.1 seconds to be relayed back to UART0
 *after 2.1sec elapsed, 1wire status ,msB and ssB send on UART0
 */
void status_com1()
{
    ringbuf_push(&comm_sndbuf, '-' );
	pushfuncpar( &srl_direct, 4, (void*)( ('+' << 8) | 'u' ) ); // will print "+uzz"

	UCSR0B |= _BV(UDRIE0); // enable transmit interrupt for UART0, so comm_sndbuf does not get stuck
	// allow 2.1 sec window for bytes coming from UART1 to be transmitted (relayed) via UART0:
	com.comm1_relay_allowed = 100; // 100 * 0.1 seconds

}

#ifdef ONE_WIRE
//http://www.vems.hu/wiki/index.php?page=ButtonImmobilizer
struct one_wire_t one_wire;

#ifndef oneWireInitLen
#define oneWireInitLen 26
// #define oneWireInitLen 14
#endif
#if oneWireInitLen==14
const uint8_t oneWireInit[oneWireInitLen] PROGMEM = {0xc1,
0xc1, 0xe1, 0x33, 0xff,  0xff, 0xff, 0xff, 0xff,  0xff, 0xff, 0xff, 0xe3, 0xc1 };
#else
// With the longer qw(c1 c1 e1 33 ff ff ff ff ff ff ff ff e3 c1 e1 33 ff ff ff ff ff ff ff ff e3 c1) double sequence it worked any number of times. I couldn't come up with anything shorter that enabled safe reread if the device any number of times..
const uint8_t oneWireInit[oneWireInitLen] PROGMEM = {0xc1,
0xc1, 0xe1, 0x33, 0xff,  0xff, 0xff, 0xff, 0xff,  0xff, 0xff, 0xff, 0xe3, 0xc1,
0xe1, 0x33, 0xff,  0xff, 0xff, 0xff, 0xff,  0xff, 0xff, 0xff, 0xe3, 0xc1 };
#endif

#ifdef ONE_WIRE_STAY_ALIVE
// do nothing with IBUTTON (or keepalive) disabled. But normally only called when IBUTTON enabled
static inline void one_wire_adc_charge(void)// set one_wire adc output to high, to charge the capacitor
{
    if( IBUTTON_WITH_KEEPALIVE ){ // both IBUTTON enabled, and keepalive also
	PORTF |= _BV(6); // charge the capacitor
	//DDRF |= _BV(6); // but only with pullup, without setting strong output
    }
}
// might also be called with IBUTTON disabled
void one_wire_adc_DIScharge(void)// set one_wire adc output to low, DIScharge the capacitor
{
    if( IBUTTON_WITH_KEEPALIVE ){ // both IBUTTON enabled, and keepalive also
	PORTF &= ~_BV(6); // DIScharge the capacitor (this only turns of pullup by itself)
	DDRF |= _BV(6);  // yes, "strong" discharge !
    }
}

/** Only called once initially, AFTER init_sensors()
* We suspect that atmega128a starts a bit sooner, at lower VCC (2.5V instead of 4V ?) which is good,
* but effects 1-wire keepalive measurement (since reference is lower, it measures higher ADC value)
* Interesting, is it such a big difference that it can give false reading?
* Yes. For example measures 5b instead of 30 (original keepalive threshold was 40)
*/
void check_1wire_adc(void)
{
    if( ! IBUTTON_WITH_KEEPALIVE ) return; // not configured for 1-wire keepalive

      // one_wire.status = ONE_WIRE_ADC_CHECK; // not needed, we examine right away:
	uint8_t adc_value = sensors8(ONE_WIRE_ADC); //check adc input if keepalive capacitor is charged
#ifdef STACKPOINTER_DEBUG	
	profiling[2] = adc_value;	// save it for debug
#endif
        if (adc_value > ONE_WIRE_ADC_LIMIT){ // capacitor is charged
            one_wire_adc_charge(); // charge further
            one_wire.status = ONE_WIRE_FOUND_CORRECT;
        }else{
            one_wire_adc_DIScharge(); // DDRF |= _BV( ONE_WIRE_ADC ); // bit6 0x40 : DDR=output (PORT=0 already) to discharge ADC out capacitor. Should NOT be needed at any other place (eg. before bootloader ?)
            one_wire.status = ONE_WIRE_LISTENING;
        }
}
#endif // ONE_WIRE_STAY_ALIVE

/** send 1-wire magic sequence via UART1
* Act (once) on ONE_WIRE_CRC_OK => ONE_WIRE_FOUND_CORRECT
* Called from userspace.
*/
void read_1wire()
{
    uint8_t s;
    if ( ! IBUTTON_ENABLED ) return;// ibutton not configured

    s = one_wire.status;
    if (s == ONE_WIRE_FOUND_CORRECT) return; //ibutton found and correct

    if (s == ONE_WIRE_CRC_OK){//1wire read success, crc ok, test if code correct
        uint8_t h,l;
        h = one_wire.ssB;
        l = one_wire.msB; // for some reaons l = msB (probably because Ibutton sends reverse)
        if ( ( config_s.prohibit_h == h) && ( config_s.prohibit_l == l)
            && IBUTTON_ANTITHEFT // cornercase: if buttoncode==fw specific prohibit code and Ibutton only used for page-B
            ){//code correct for button 1
              s = ONE_WIRE_FOUND_CORRECT;
           } else if((config_s.ibutton_code2[0] == h && config_s.ibutton_code2[1] == l) ){
                // correct but load eeprom page B (page 1) !
                load_variables_wrapper(1);
                s = ONE_WIRE_FOUND_CORRECT;
          }else{
                s = ONE_WIRE_FOUND_INCORRECT;//crc ok, but code incorrect, need it for testing ibutton and show it
          }

            one_wire.status = s;
            IgnInjProhibitCheck(); // likely needed at least before fuelpump_on() to activate output (hopefully no ill sideeffects)

          if( s == ONE_WIRE_FOUND_CORRECT){
          #ifdef ONE_WIRE_STAY_ALIVE
              one_wire_adc_charge();
#endif // ONE_WIRE_STAY_ALIVE
                // we know: old one_wire.status != ONE_WIRE_FOUND_CORRECT (see condition above)

              etime.now_upper=0; // so that config.pump_on_mintime can take its time
              engine.status_output &= ~_BV(STATUS_OUT_FUEL_BIT);  // also necessary to start fuelpump
              fuelpump_on(); //  one_wire.status  ALREADY saved, so it will work
          }
          return;
    }

    if (one_wire.tx_cnt != 0) return;
    one_wire.status = ONE_WIRE_LISTENING;
    one_wire.tx_cnt=1;
    one_wire.rx_cnt=0;
    UDR1 = pgm_read_byte( &(oneWireInit[0] ));//send the first byte
    UCSR1B |=_BV(UDRIE1);// enable data register empty interrupt on usart1

/* old busy-wait based method
	for(i=0; i< oneWireInitLen; i++){
		while( (UCSR1A&0x20) == 0 )  ;  // UDRE1 bit means: TX-Data Register Empty

		UDR1 = pgm_read_byte( &(oneWireInit[i] ));
	}
*/
}

#endif //ONE_WIRE

/***************************************************************************/
/***************************************************************************/
/*** MegaTune compatibility functions                                    ***/
/***************************************************************************/
/***************************************************************************/

uint8_t sendSignature(uint8_t i) {
  if (i < 18) { // TODO: check sizeof ?
        return pgm_read_byte( (uint8_t *)&vems_string + i );
  }
  i -= 18;
  if (i < 6)
        return pgm_read_byte( (uint8_t *)&f_version + i );
  switch (i)
  {
  case 7:	// send last
	  return MY_MAKE_BYTE_0;
  case 6:	// send first
	  return MY_MAKE_BYTE_1;
  default:
	  return 0;
  }
}

/***************************************************************************/
/*** codeVer                                                             ***/
/***************************************************************************/
// return codeversion "VEMS v1.1 12x12 kpa=2,1"
uint8_t MTcodeVer(uint8_t i) {
  switch(i){
  	case 0: return 'V';
  	case 1: return 'E';
  	case 2: return 'M';
  	case 3: return 'S';
  	case 4: return ' ';
  	case 5: return 'v';
  	case 6: return '0' + FIRMWARE_VERSION_MAJOR;
  	case 7: return '.';
  	case 8: return '0' + FIRMWARE_VERSION_MINOR;
  	case 9: return '.';
  	case 10: return '0' + FIRMWARE_VERSION_BUILD / 10; // TODO: make sure gcc optimizes this constant
  	case 11: return '0' + FIRMWARE_VERSION_BUILD % 10;
  	case 12: return ' ';
  	case 13: return '0' + VE_SIZE_RPM / 10; // TODO: make sure gcc optimizes this constant
  	case 14: return '0' + VE_SIZE_RPM % 10;
  	case 15: return 'x';
  	case 16: return '0' + VE_SIZE_KPA / 10; // TODO: make sure gcc optimizes this constant
  	case 17: return '0' + VE_SIZE_KPA % 10;
  	case 18: return ' ';
  	case 19: return 'k';
  	case 20: return 'p';
  	case 21: return 'a';
  	case 22: return '=';
  	case 23: return KPAFAC_UNIT == 0xF ? '2' : hex_hi(config.kpa_conf);
  	case 24: return ',';
  	case 25: return KPA_CONFIG_RESOLUTION == 0xF ? '1' : hex_lo(config.kpa_conf);
	default: return 'V';
    }
#if(MT_CODEVER_LEN != 26)
#warning "MT_CODEVER_LEN size mismatch"
#endif

}

/***************************************************************************/
/*** testComm                                                            ***/
/***************************************************************************/
// return low part of second-count
uint8_t MTtestComm(uint8_t i) {
  return (rtc.sec & 0xFF);
}



/***************************************************************************/
/*** getMCPdataHIGH                                                      ***/
/***************************************************************************/
// does not start a adc conversion, just masks out the high byte of latest conversion
// and returns if the channel data is not 0xffff

// channels range from 0-7 internally (from 1 to 8 according to datasheet)

uint8_t getMCPdataHIGH(uint8_t i) {
	uint16_t t;
	uint8_t r;

	t = get_mcp3208_data12( i );
	r = t >> 8;
	return r;
}


/***************************************************************************/
/*** getMCPdataLOW                                                       ***/
// Normally called from userspace.
// Starts an mcp channel ADC conversion on the appropriate channel
// and if conversion is finished return low byte of adc result, else 0
// channels range from 0-7 internally (from 1 to 8 according to datasheet)
/***************************************************************************/
uint8_t getMCPdataLOW(uint8_t i) {
	uint16_t t;
	uint8_t r;

	t = get_mcp3208_data12(i);
	r = (uint8_t) t;
#ifndef MAINLOOP_TASKER //not needed for tasker
	read_mcp3208(i);
#endif
	return r;
}

/*
* Sends the data with byte stuffing, but without heading
* needed for config table stuffed dump
* SendPage_stuffed must call after this
*/
uint8_t SendPage_data_stuffed (uint8_t cnt)
{
	//cmdqueue[head].private must have pageID
    ringbuf_push_bytestuff_updatecrc(&comm_sndbuf,send_page(cnt));
    return 'D';

}
/* sends the selected page in triggerlog format, with address or page name
format: N byte data + 2 byte address ( or 0xff+page ID) + 1byte SEND_PAGE_STUFFED_TYPE(119) + 2 byte CRC + HDLC_END
*/
uint8_t SendPage_stuffed( uint8_t i )
{
    uint8_t size = cmd_queue[cmd_queue_head].len -1 ;
    uint8_t revese_cnt= size-i;

	uint8_t v;
	if(i>=(size)){
		ringbuf_push_HDLC_END_commcrc_ffff(); /* send framing char, do NOT stuff*/
	} else if(revese_cnt>=3){/* data section with address and trtype*/

        uint16_t addr_or_page = (uint16_t)cmd_queue[cmd_queue_head].private;
	    if ( revese_cnt > 5){
            uint8_t *address = (cmd_queue[cmd_queue_head].private);
            if (address != NULL) v = address[i];
            else v = SENDPAGE_ERROR_NO_PAGEONID;
	    } else if ( revese_cnt == 3 ) v = READ_SRAM_STUFFED_TYPE_COMMAND-128;//triggerlog type
	    else if ( revese_cnt == 4 ) v = (uint8_t)addr_or_page ;
	    else v = addr_or_page >> 8;

	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else { /* i == size-3 or size-2*/
		v = (i == (size-2) ) ? com.commcrc & 0xff : com.commcrc >> 8;/* CRC sent in little endian as in SIPR*/
		ringbuf_push_bytestuff(&comm_sndbuf, v); /* When sending crc itself => doNT update crc*/
	}

    return 'P';
}

/** sends RT data in triggerlog format (sometimes pushes 2 bytes because of HDLC_ESC)
*  MT_RTVAR_COUNTMORE byte data+ 1 bytetriggerlog type(32) + 2 byte crc + 1 byte HDLC_END
*/
uint8_t MTsendRTvar_triggerlog(uint8_t i)
{
    uint8_t v;
    if( i <= MT_RTVAR_COUNTMORE ){
        if (i == MT_RTVAR_COUNTMORE) v = RT_STUFFED_TYPE;
        else{
            v = MTsendRTvar(i, 0);
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( i >= (MT_RTVAR_COUNTMORE + 3) ){ // MT_RTVAR_COUNTMORE +3
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
        } else { // MT_RTVAR_COUNTMORE +1 or +2
		v = (i == (MT_RTVAR_COUNTMORE + 1) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 0;
}

#if MODBUS_ON_COMM == 2
// extended array: 1: &com.atomic_helper 2: &com.atomic_helper+1, ....
#define ALTERNATIVE_ATOMIC_HELPER_PTR ((&com.atomic_helper-1) +caller)
#else   // normally
#define ALTERNATIVE_ATOMIC_HELPER_PTR (&(com.atomic_helper))
#endif // not MODBUS_ON_COMM == 2

// could save a few flashbytes by consuming +1 SRAM byte and always use global var com.atomic_helper/com.atomic_helper_more[]
#define ATOMIC_HELPER_SRL &cmd_queue[cmd_queue_head].privch.a
#define ATOMIC_HELPER (uint8_t *)(caller==0 ? ATOMIC_HELPER_SRL : ALTERNATIVE_ATOMIC_HELPER_PTR )

/**
* i: index 0..len-1
* if caller==0 store ATOMIC_HELPER in cmdqueue,
* caller==1 com.atomic_helper global var (SDcard)
* caller==2 com.atomic_helper+1 global var (2nd serial port)
*/
const uint8_t PROGMEM COMM_AD_CHANNELS[8]={4,3,6,7,5,0,1,2};//send and ad order does not match


// desired, no jump tables for the rest of the code, but jump table for comm.c MTsendRTvar, cant get it to work, gcc please behave :) - DB. 
// #pragma GCC optimize ("freestanding", "no-inline-small-functions", "short-enums", "unsigned-char", "unsigned-bitfields", "pack-struct", "no-caller-saves")

/** i=0 .. MT_RTVAR_COUNTMORE-1,  caller= ... see above
* Must be called from userspace, because of sei() ... SREG=sreg_bkup would be more correct
* is earliest sei() at i==52, or also hidden somewhere earlier ? */
uint8_t MTsendRTvar(uint8_t i, uint8_t caller) {

switch (i) {
  case 0: return (uint8_t)rtc.sec;  //secl
  case 1: return engine.boostalt_command_duty;
  case 2: return engine.status;  //engine
  case 3: return engine.kpa_or_tps;  // AlphaN load; originally sensors8(BARO)
  // high resolution MAP
  case 4: {
      *ATOMIC_HELPER = engine.kpa250 & 0xFF;
      return engine.kpa250 >> 8;
  }
  case 5: return *ATOMIC_HELPER;
  case 6: return engine.mat;   //mat
  case 7: return engine.coolant;   //clt
  case 8: return engine.tps;   //tps
  case 9: return engine.batt;  //batt
  case 10: // o2
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED)
	{
		// traditional stuff via lookup table: return wbo2.Ip_output;
		return runtime_lambda(0);
	}
	else
#endif
		return ego_voltage();  //ego
  case 11: return mult16_8(corr.ego16,100)>>8;	//egocorr
  case 12: 
	if(isInjProhibited()) 
		return corr.air; 
	else {
#ifdef WBO2	
		if (WBO2_SENSOR_CONFIGURED)
			return runtime_lambda(1); //aircorr actually NEVER sent. Either Nissan maxwinlen or lambda
		else
#endif		
			return ego_voltage2();  //ego 2 voltage
	}		
  case 13: return (corr.warm >=  1535) ? 255 : (mult16_8(corr.warm-230,50)>>8);      //warmcorr railed to 600%


#ifdef CAN_MCP2515
  case 14: {
	if(ETC_CONFIGURED && ETC_MSS_IDLE_ETC_CAN) {
		switch(ETC_MSS_IDLE_ETC_RT) {
			default:
			case 0: *ATOMIC_HELPER = engine.rpm_vhr & 0xFF; return engine.rpm_vhr >> 8;
			case 1: *ATOMIC_HELPER = idle_etc.act_class[ETC].act[0].pos & 0xff; return idle_etc.act_class[ETC].act[0].pos >> 8;
			case 2: *ATOMIC_HELPER = idle_etc.act_class[ETC].act[1].pos & 0xff; return idle_etc.act_class[ETC].act[1].pos >> 8;
			case 3: *ATOMIC_HELPER = idle_etc.act_class[IDLE].act[0].pos & 0xff; return idle_etc.act_class[IDLE].act[0].pos >> 8;
			case 4: *ATOMIC_HELPER = idle_etc.act_class[IDLE].act[1].pos & 0xff; return idle_etc.act_class[IDLE].act[1].pos >> 8;
			case 5: *ATOMIC_HELPER = idle_etc.act_class[ETC].act[0].tgt & 0xff; return idle_etc.act_class[ETC].act[0].tgt >> 8;
			case 6: *ATOMIC_HELPER = idle_etc.act_class[ETC].act[1].tgt & 0xff; return idle_etc.act_class[ETC].act[1].tgt >> 8;
			case 7: *ATOMIC_HELPER = idle_etc.act_class[IDLE].act[0].tgt & 0xff; return idle_etc.act_class[IDLE].act[0].tgt >> 8;
		}
	} else {	
		*ATOMIC_HELPER = engine.rpm_vhr & 0xFF; 
		return engine.rpm_vhr >> 8;  
	}	
  }
#else
  // high resolution RPM
  case 14: {
    *ATOMIC_HELPER = engine.rpm_vhr & 0xFF; 
	return engine.rpm_vhr >> 8;  
  }
#endif
 
  case 15: return *ATOMIC_HELPER;

//show O2 instead pw in vemstune for mde02 or mde03 sensor calibration with stalled engine
  case 16: {
	uint16_t t16;
	if((engine.rpm==0) && (debug.enable_flags & _BV(1)) // mde02 (calibrating WBO2 sensor1) or mde03 (calibrating WBO2 sensor2)
        && ((t16 = (debug.enable_flags & _BV(0) ? WBO21.current_o2 : WBO20.current_o2)) != 0)) {
		; // t16 = already done ...
	} else {
		t16 = engine.calculated_fuelpw;
	}
	*ATOMIC_HELPER = t16 & 0xFF;
	return t16 >> 8;     //pw high
  }
  case 17: return *ATOMIC_HELPER;   //pw low
  case 18: return corr.baro;      // barocorr is 128 based
  case 19: {	// calculate from the stored high-res value (/100 and railing is needed)
		uint16_t ge;
		ge = mult16_8(corr.gammae381, 68) >> 8;// 381 * 68 / 256 = 101 (NOT 100 !)

		if ((ge >> 8) != 0)  // check for overflow
			return 0xFF;
		else
			return (uint8_t)ge;
	}
  case 20: return corr.ve_hp;
  case 21: return engine.dwell>>4;	// 64 usec resolution (possibly but unlikely overflowing)
  case 22: return (engine.ignadv < 510 ? (engine.ignadv>>1) : 255); // 128 + ignadv_halfdeg
  case 23: return iac.position;
  
  case 24:
  case 25:
  case 26:
  case 27:
  case 28:
  case 29:
  case 30:
  case 31:
  case 32:
  case 33:
  case 34:
  case 35:
  case 36:
  case 37:
  case 38:
  case 39:
 	
	// helper for analog value send in the historic order, todo send in real order
	// ************************************************************************************************
	// why not descramble this strange mcp channel order ?? we can easily edit vemtune ini to match - DB
	// ************************************************************************************************
	{ // 24 to 39
        uint8_t channel;
#ifdef KNOCK_DEBUG // EVIL, steals the last 4 mcp3208 channels (although Dave made indications of this in VT)
        if ((i >= 32) && alternative_rtdata()) return (KNOCK_RETARD_ALL ? knock_cyl[0].knock_adjust : knock_cyl[i-32].knock_adjust); // knock_cyl[0] to [7]
#endif
        i -= 24;
        channel = pgm_read_byte( &COMM_AD_CHANNELS[i >> 1]);//get the historic order
        if(i & 0x01){
            return *ATOMIC_HELPER;
        } else { // even index first, store the following second(odd) byte
#ifndef MAINLOOP_TASKER		
			// read_mcp needed in getMCPdataLOW
			*ATOMIC_HELPER = getMCPdataLOW(channel);
            return getMCPdataHIGH(channel);
#else

#ifdef INPUT_MUX_USED	
			uint8_t muxs = (get_comm_mux_state() << 1);	
			if(muxs) { // non zero mux state meaning muxed channels might need store
				if(channel == input_mux_lowch()) {
					channel = 6 + muxs; // even channels -> 8, 10, 12
				} else if (channel == input_mux_highch()) {
					channel = 7 + muxs; // odd channels -> 9, 11, 13
				} 
			}
#endif
			// read_mcp done in mainloop, no chance for data change to sneak in between, no 0xffff check can only be 0xffff on first (non) sample after powerup.
			uint16_t mcp_data = get_mcp3208_data12(channel);
			*ATOMIC_HELPER = (uint8_t) mcp_data;
			return mcp_data >> 8;
#endif
    	}
	} 
  // make sure 16 bit variables start on even offsets:
  case 40: { 
#ifdef INPUT_MUX_USED 
	if (input_mux_config_enabled()) set_next_comm_mux_state(); 
	else reset_comm_mux_state();
#endif	
	return corr.lambda; 
  }

#ifdef A340E // andrey always use alt_knock
  case 41: return at_control.at_temp;
#else
  case 41: return knock.worst_knocking_val; // knock value of the WORST knocking cyl
#endif

#ifdef SECOND_VSS
  case 42: return wheel_speed2.wspeed;
#endif // SECOND_VSS

  case 43: return knock_pack(7); // cylinder specific knock, 2 bit/cyl 7..4
  case 44: return knock_pack(3); // cylinder specific knock, 2 bit/cyl 3..0

  case 45: {
      uint8_t wka = KNOCK_RETARD_ALL ? knock_cyl[0].knock_adjust : knock.worst_knock_adjust; // knock.worst_knock_adjust could be good in any case, but...
      return wka;  // knock adjust for the WORST knocking cyl
  }
  case 46: return mmc.timer; // TEMPORARY variable... might change. roughly: >= 248 means SD logging is not active

  // status bytes like WBO2 activated and alike
  case 47: {// bit1 unused later used for trigger error, no it send at byte 66
	uint8_t status = 0;
#ifdef ALS
	if ( is_als_switch_activated() ) status |= 0x80;	// als switch active (ALS function active is in byte 52)
#endif
	if( n2o_is_function_active() ) status |= 0x01; // N2O active
#ifdef SOFTLAUNCH
	if ( is_launch_input_or_function_active() ) status |= 0x40; // launch control active or launch input button active
#endif
	if ( during_actual_shiftcut() ) status |= 0x20; // shift cut
	if ( is_injector_staging_active() ) status |= 0x10; // inj stage2 active
	if (engine.status4 &_BV(STATUS4_CONFSW) ) status |= 0x08;	// Config slot
	if (isO2_closedloop()) status |= 0x04; // closed loop (ego)
	if (eeprom.flags & _BV(active)) status |= 0x02; // eeprom saving in progress flag. Since 1.1.83
	return status;
  }
#ifdef BOOSTCONTROL
	// without boostcontrol the default 0 will be returned
	case 48: return boost_data.boost_target16 >> 5; // 1/8 kPa => 4 kPa
	case 49: return engine.boost_command_duty; // boost solenoid duty
	case 50: return (uint8_t)((boostcontrol_pid.integral+0x4000)>>7); // boost PID integral, assume PID_INTEGRAL_EXP=6
#endif
	case 51: return (uint8_t)(iac.integral>>8)+0x80; // iac.integral, assume INTEGRAL_LIMIT_SCALE=7
	case 52: { // bit5=als function active;  bit4:0=engine.sparkcutlvl
	    uint8_t retval = engine.sparkcutlvl; // 0..16 (0 is no igncut, normal operation)
	    cli(); clear_individual_allowed(); sei(); // unrelated, prepare for byte53
	    if( retval > 16 ) retval = 31;  // error indication only, should never happen
	    retval |= is_als_active(); // bit5: is_als_active() ? _BV( als_active ) : 0
		retval |= engine_is_knocking() ? 64 : 0; // bit6: when knock detected, this bit is true even when no retard is applied
        return retval;
	}

#if 0 // old style two choice data
	case 60: {
		cli(); 
		set_individual_allowed(); 
		sei();
	}	
	case 53:
	case 54:
	case 55:
	case 56:
	case 57:
	case 58:
	case 59: {
		uint8_t channel = i - 53;
		return (alternative_rtdata()) ? knock_val8(channel) : individual_power[channel];
	}
#else // new style 3 choice data

// TESTING TODO -> REMOVE	
#ifndef MAINLOOP_RUNTIME_SCOPE // not MAINLOOP_RUNTIME_SCOPE
	case 53: { 
		if(alternative_rtdata()) return knock_val8(0); 
		if(engine.rpm) return individual_power[0]; 
		return (debug.enable_flags & _BV(0)) ? WBO21.status : WBO20.status;
	}	
	case 54: { 
		if(alternative_rtdata()) return knock_val8(1); 
		if(engine.rpm) return individual_power[1]; 
			
		// RPM=0, send wbo2.Ri
        // mde02 (calibrating WBO2 sensor1) or mde03 (calibrating WBO2 sensor2):
        uint16_t t16 = (debug.enable_flags & _BV(0)) ? WBO21.Ri : WBO20.Ri;
        *ATOMIC_HELPER = t16 & 0xFF;
        return t16 >> 8; // send high byte first
	}	
	case 55: { 
		if(alternative_rtdata()) return knock_val8(2); 
		if(engine.rpm) return individual_power[2];  
		return *ATOMIC_HELPER;
	}	
	case 56: { 
		if(alternative_rtdata()) return knock_val8(3); 
		if(engine.rpm) return individual_power[3];  
		
		// RPM=0, send wbo2.nernst_dc WILL be reused later. (See case 66 bit3:0)
        // mde02 (calibrating WBO2 sensor1) or mde03 (calibrating WBO2 sensor2):
        uint16_t t16 = (debug.enable_flags & _BV(0)) ? WBO21.nernst_dc : WBO20.nernst_dc;
        *ATOMIC_HELPER = t16 & 0xFF;
        return t16 >> 8; // send high byte first
	}	
	case 57: { 
		if(alternative_rtdata()) return knock_val8(4); 
		if(engine.rpm) return individual_power[4];  
		return *ATOMIC_HELPER;
	}	
	case 58: { 
		if(alternative_rtdata()) return knock_val8(5); 
		if(engine.rpm) return individual_power[5];  
		return (debug.enable_flags & _BV(0)) ? WBO21.heat : WBO20.heat;
	}	
	case 59: { 
		if(alternative_rtdata()) return knock_val8(6); 
		if(engine.rpm) return individual_power[6];  
		return (debug.enable_flags & _BV(0)) ? WBO21.pump_pw : WBO20.pump_pw;
	}	
	case 60: {
		cli(); 
		set_individual_allowed(); 
		sei();
		return (alternative_rtdata()) ? knock_val8(7) : individual_power[7];
	}	
#else // MAINLOOP_RUNTIME_SCOPE
	case 53: return WBO20.status;
	case 54: {
        uint16_t t16 = WBO20.Ri;
        *ATOMIC_HELPER = t16 & 0xFF;
        return t16 >> 8; // send high byte first
	}	
	case 55: return *ATOMIC_HELPER;
	case 56: { 
		uint16_t t16 = WBO20.nernst_dc;
        *ATOMIC_HELPER = t16 & 0xFF;
        return t16 >> 8; // send high byte first
	}	
	case 57: return *ATOMIC_HELPER;
	case 58: return WBO20.heat;
	case 59: return WBO20.pump_pw;
	case 60: {
		cli(); 
		set_individual_allowed(); 
		sei();
		return (alternative_rtdata()) ? knock_val8(7) : individual_power[7];
	}	
#endif // MAINLOOP_RUNTIME_SCOPE	
	
#endif // eof new style 3 choice 
	
	case 61: return engine.tok_diff; // this dV/dt will stay
#ifndef  PS2KEYBOARD
	case 62: return wheel_speed1.wspeed;  //speed
	case 63:
	{
#ifdef A340E		
		uint8_t t8=0;

		if (at_control.at_status&_BV(LOCKUP))
			t8=100;
		return engine.gear+t8;	//gear
#else
		return engine.gear;	//gear
#endif		
	}
#else
	case 62: return now_64msec() - engine.launchinact_time;  //elapsed time
	// case 63: // empty in non wheelspeed
#endif
#ifdef CAMSHAFT_ANGLE
#ifndef FOUR_CAMSHAFT_CONTROL
    case 64: return cam.data[INTAKE_CAM1].position;
    case 65: return (cam.data[INTAKE_CAM1].duty.single >> 4); // camshaft intake control duty cycle, logging for dual output crippled, rethink this - DB
#else
	case 64: return CAM_RT_USE_SECOND_SET ? cam.data[INTAKE_CAM2].position : cam.data[INTAKE_CAM1].position;
    case 65: return CAM_RT_USE_SECOND_SET ? (cam.data[INTAKE_CAM2].duty.single >> 4) : (cam.data[INTAKE_CAM1].duty.single >> 4); 
#endif	
#endif
    case 66: //status2 (trigger error bits (bit 4:6). bit 3:0 TEMPORARILY ri_absdiff. bit7 = WBO2 NOT ready
    {
        uint8_t status = WBO20.nernst_absdiff>>4; // bit 3:0 nernst_absdiff. Maybe TEMPORARY (7..8 are best midvalues);
        if (engine.trigger_error & 0x0f) // lower 4 bit need to turn on error for 1.6sec if zero error code invalid
            status += (engine.trigger_error & 0x70 ); // bit 7 reserved for future trigger error code
        if(engine.status_ext & _BV(o2_not_ready)) status += 0x80; // WBO2 ri or nernst off-target
        return status;
    }
#ifdef CAMSHAFT_ANGLE_EXHAUST
#ifndef FOUR_CAMSHAFT_CONTROL
    case 67: return cam.data[EXHAUST_CAM1].position; // third trigger angle
    case 68: return (cam.data[EXHAUST_CAM1].duty.single >> 4); // camshaft exhaust control duty cycle, logging for dual output crippled, rethink this - DB
#else
	case 67: return CAM_RT_USE_SECOND_SET ? cam.data[EXHAUST_CAM2].position : cam.data[EXHAUST_CAM1].position; 
    case 68: return CAM_RT_USE_SECOND_SET ? (cam.data[EXHAUST_CAM2].duty.single >> 4) : (cam.data[EXHAUST_CAM1].duty.single >> 4); 
#endif	
#endif
    case 69: return engine.status_output;
    case 70: return engine.partialfuelcut;
    case 71: return mult16_8(corr.ego16_2, 100)>>8;	// egocorr
    case 72: return (corr.tpsaccel >= 2040) ? 0xFF: (corr.tpsaccel>>3);  // tpsaccel, 4usec => 32 usec ;
    case 73: 
	{ /*{ // send pw2 ... good for debugging, might be replaced later with something more useful
        uint16_t t16 = engine.calculated_fuelpw2; // *ATOMIC_HELPER = t16 & 0xFF;
        t16 <<= 3;  // upper byte is 1024 => 128 usec (lower byte was 4 usec). 128 usec resolution (0-32 msec range)
        return t16 >> 8;     // pw2 MSB
        }*/ 
		return iac.target_rpm;
	}
    case 74: 
	{	
		// what does this do? auto dump of comm realtime values with mde80 (auto comm dump of rt data).
		if (debug.enable_flags & _BV(7)) {
			pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNTMORE);
		} 
		// not needed we can calc in vemstune without any problem, data below not used in VT
		// find something more usefull to report - DB
		// return ENGINE_ETC_TARGET8(); // (engine.ETC_piddata.target <<2) >> 8
		return get_comm_mux_state();
	}
	default: return 0;
  }
}

// #pragma GCC reset_options



/** send 32 bytes:
 * first 16 bytes is the 8 mcp3208 in big endian
 * 2nd 16 bytes is the 7 internal AD in big endian, but channel 7 first (0 latest)
 */
uint8_t allSensorRaw(uint8_t i)
{
	uint8_t t8, *p;
	if( i < 16 ){ // 0 .. 15
		t8 = i >> 1;
		if( i & 1 ){ // LOW sent AFTER high byte
			return getMCPdataLOW( t8 );
		} else { // send high FIRST:
			return getMCPdataHIGH( t8 );
		}
	}
	p = (uint8_t*)sensor_filt; // convert to byte array
	return p[ 31 - i ];
}

/* Dump the table
* private higher byte holds the table identifier char,
* private lower byte holds the table-line
* line higher 4 bits contain the counter (eg 0..7)
* line lower 4 bits contain the maxvalue of counter (eg 7)
* count (conservatively) holds the character inside the line
*/
uint8_t TableDump(uint8_t cnt){
	uint16_t t;
	char table;
	uint8_t table_lines; // number of table lines without the r[] k[] lines.
	uint8_t line; // 0..15, counts up after every line. if(line > table_lines-1) than .[r]=.. or .[k]=.. printed
	uint8_t linemax; // 0..15, stays same determined by menu.c TableDump "caller"/initializer
	uint8_t *p;
	uint8_t len; // elements/line. Higher than 16 is allowed (if line length is less than 240 characters)
	uint8_t line_end; // calculated: HEADER_LENGTH + len * block_printlen; Must be less than 240, ~len<78

// if 2 is allowed, than allow case 3,4 in the last switch-case
#define block_bytes 1
#define block_printlen 3

	t = (uint16_t)(cmd_queue[cmd_queue_head].private);
	table = t>>8;

	if(cnt >= 0xF0){
		// similar to state 0, but 0 cannot be used again for some reason...
		cmd_queue[cmd_queue_head].count = 0;
		return table;
	}

	line = t & 0xFF;
	linemax = line & 0x0F;
	line >>= 4;
	// block_printlen = 3;
	// block_bytes = 1;

	table_lines = 0xFF; // this way line > table_lines condition not triggered

	switch(table){ // this should get from a place common with menu
		case 'j':
			p=ve.ve8_table;
			len = VE_SIZE_RPM;
			table_lines = VE_SIZE_KPA;
			break;
		// case 'k':
		//	p=ve.kpa_range;
		//	len = VE_SIZE_KPA;
		//	break;
		case 'l':
			p=ve.lambda;
			len = LAMBDA_SIZE_RPM;
			table_lines = LAMBDA_SIZE_KPA;
			break;
		// either ifndef NOIGN is not needed, or needed in menu.c too
		// I allow the read-write of 'n' table for other applications...
		case 'n':
			p=ign.adv_table;
			len = IGNADV_SIZE_RPM;
			table_lines = IGNADV_SIZE_KPA;
			break;
#ifdef HWMAP
		case 'h':
			p=hwmap.table;
			len = HWMAP_SIZE;
			break;
#endif

#ifdef A340E
		case 'u':
			p = at_config.shift_up[0];
			len = sizeof(at_config);
			break;
#endif

#ifdef TOYOTA_1JZ_VVTI
		case 'z':
			p = vvti_map.vvti;
			len = VE_SIZE_RPM;
			break;
#endif

		case 'f':
			p = (uint8_t*)&knock_table;
			len = KNOCK_SIZE_RPM;
			break;

#ifdef BOOSTCONTROL
		case 'b':
			p = boost.boostmaxtarget_table;
			len = BOOST_SIZE_RPM;
			table_lines = BOOST_BTABLE_GEARLINES;
			break;
		case 't':
			p = boost.tpstransfer_table;
			len = TPSTRANSFERTABLE_SIZE;
			break;
		case 'y':
			p = boost.boost_ref_table;
			len = IGNADV_SIZE_RPM;
			table_lines = BOOST_SIZE_KPA;
			break;
#endif
#ifdef MAT_ENRICH
		case 'e':
			p = ve.mat_enrich;
			len = 8;
			table_lines = LAMBDA_SIZE_KPA;
			break;
#endif
#ifdef CAMSHAFT_ANGLE
#ifdef TOYOTA_1JZ_VVTI
#error "CAMSHAFT_ANGLE incompatible with TOYOTA_1JZ_VVTI"
#endif
		case 'z':
			p = ve.camshaft_angleI;
			len = 16; // theoretically 16+2 lines with 8 in each, but hack-sent this way
			table_lines = 8;
			break;
#endif
#ifdef CHARGE_TEMP
		case 'e':
			p = f_charge.f_charge;
			len = VE_SIZE_RPM;
			table_lines = VE_SIZE_KPA;
			break;
#endif
#ifdef MAP_CORR
		case 'i':
			p = ve.map_corr;
			len = MAP_CORR_SIZE_RPM;
			table_lines = MAP_CORR_SIZE_KPA;
			break;
#endif
		case 's':
			p = ve.launch_rpmtable;
			len = 8;
#ifdef CLT_IGN_ADJUST			
			// was table_lines = 2 dirty hack to dump clt_ign_adjust + clt/kpa_range;
			table_lines = 2+3;
#else
			table_lines = 2;
#endif			
			break;
		case 'p':
			p = boost.anytrim1_array;
			len = 17; // 3*17 bytes
			table_lines = 3;
			break;
		case 'w':
		default:
			p=ve.mat_retard;
			len = LAMBDA_SIZE_RPM;
			table_lines = LAMBDA_SIZE_KPA;
			break;
	}

#define HEADER_LENGTH (5)
	if(cnt < HEADER_LENGTH){ // line beginning
		switch(cnt){
			case 0: return table;
			case 1: return '[';
			case 2: if( line < table_lines ){ // print kpa line as j[k]=xx xx xx ...
					return hex_lo(line);
				} else if( line == table_lines ){
					return 'r';
				} else return 'k';
			case 3: return ']';
			case 4: return '=';
		}	// switch
	}	// if

	// we must determine how long the line is for the given table:
	// 3 * len or 5 *len for a 16 bit table
	// (-1 in any case, since no space before first, see the '\r' branch)
	if( table == 'j' && VE_SIZE_KPA == 14 && line == 13+2 ){
		// for 16x14 j table last (kpa) line is 14 bytes only
		line_end = 14;
	} else line_end = len;
	line_end *= block_printlen;
	line_end += HEADER_LENGTH;

	if(cnt == line_end){ // shell we finish or go to next line
		if(line != linemax){ // not finished
			// endianess dependent and bigger flash: uint8_t *little_endian_p; little_endian_p = &cmd_queue[cmd_queue_head].private; * little_endian_p += 16;
			// line++;
			// solution that does not assume little endian:
			cmd_queue[cmd_queue_head].private += 16; // = (void*)((table<<8) + (line<<4) + linemax);
			cmd_queue[cmd_queue_head].count = 0xF0-1; // will be incremented (0-1 doesnt work)
		} else {
			if( table == 'e'){//this condition + call adds +178 byte to flash !?!
				// 0xF240 - 0xF18E = 178 = 0xF6C4 - 0xF612 (with either my_make)
				// moving pushfuncTableDump() above did not help either
				pushfuncTableDump( TABLES_TO_DUMP2, sizeof(TABLES_TO_DUMP2) >> 1 );
			}
			cmd_queue[cmd_queue_head].count = 0xFE; // set to len => THE END
		}
		return '\n';
	}

	if(cnt == line_end - 1){
		return '\r';
	}

	cnt -= (HEADER_LENGTH - 1);

	p += len * line * block_bytes;

	// calculate modulus without division:
	while(cnt >= block_printlen){
		cnt -= block_printlen;
		p += block_bytes;
	}

	switch(cnt){	// every byte is printed to 3 characters
		// case 0: return ' '; // handled by default branch
		case 1: return hex_hi(*p);
		case 2: return hex_lo(*p);
#if(block_bytes>1)
		case 3: return hex_hi(*(p-1));	// only for 16 bit table (little endian)
		case 4: return hex_lo(*(p-1));
#endif
		default: return ' ';
	}
}

/* Dump the config structure:
 * We keep track of line number in private.
 * position inside the line is in count, but somewhat tricky
 * Note that len is 0xFE because of pushfunc() hack, although we specified 0xFF
 */
uint8_t ConfigDump(uint8_t cnt) {
	uint8_t c, ch, i, *str;
	uint8_t last_element;
	uint16_t idx;
	PGM_VOID_P P;
	// not always needed, but almost:
	// if(cnt == 0xF7) cmd_queue[cmd_queue_head].count = cnt = 0; // needed because of the len-1 hack

	idx = (uint16_t)(cmd_queue[cmd_queue_head].private);
	ch = cmd_queue[cmd_queue_head].privch.a;
	if ( ch == 'C') {
	    P = (PGM_VOID_P)(varstr_ext + idx);
	    last_element = (sizeof(struct config_ext_t)-1);
	}
	else if( ch == 's') {
	    P = (PGM_VOID_P)(varstr_shared + idx);
	    last_element = (sizeof(struct config_shared_t)-1);
	}else{
	    P = (PGM_VOID_P)(varstr + idx);
	    last_element = (sizeof(struct config_t)-1);
	}
	i = 0;
	while((str=(void*)PRG_RDW(P)) == NULL){
		P -= 2;  // addresses are two-byte
		i++;
	}

// #define LAST_ELEMENT (sizeof(struct config_t)-1)

	if(cnt < 0xF8) {
		c = PRG_RDB(str + cnt);
		if(c) {
			return c;
		} else { // end of string
		// we need the [.]= if i>0 or the next pointer is NULL but not at the very end
			if(i || (((void*)PRG_RDW(P+2)) == NULL && idx < last_element)){ // no need for [i]=.. just =..
				cmd_queue[cmd_queue_head].count = 0xF8 - 1 ; // will be incremented!
				return '[';
			} else {
				cmd_queue[cmd_queue_head].count = 0xFB - 1 ; // will be incremented!
				return '=';
			}
		}
	}

    if (ch == 'C') str = ((uint8_t *)(&config_ext));
    else if (ch == 's') str = ((uint8_t *)(&config_s));
	else str = ((uint8_t *)(&config));

	switch (cnt){
		case 0xF8: return hex_ch(i);
		case 0xF9: return ']';
		case 0xFA: return '=';
		case 0xFB: return hex_hi(*((uint8_t *)str + idx));
		case 0xFC: return hex_lo(*((uint8_t *)str + idx));
		case 0xFD: return '\r';
		case 0xFE: // let's see: are we at the very end ?
			if((uint16_t)(++ cmd_queue[cmd_queue_head].private) <= last_element ){
				// cmd_queue[cmd_queue_head].count = 0xFE;	// end of the show
				// = cmd_queue[cmd_queue_head].len
				cmd_queue[cmd_queue_head].count = 0xFF; // show must go on next line
			};
			return '\n';
	}
	return 'z';
}

/***************************************************************************/
/***************************************************************************/
/*** End of MegaTune compatibility functions                             ***/
/***************************************************************************/
/***************************************************************************/




/***************************************************************************/
/*** sendConfigVar                                                       ***/
/***************************************************************************/
uint8_t sendConfigVar(uint8_t i) {
  if (i < sizeof(struct config_t))
    return *( (uint8_t *)&config + i );
  else
    return 0;
}


/*
 * returns upper and lower byte of "private"
 */
uint8_t srl_direct( uint8_t i ){
	uint16_t t = (uint16_t) cmd_queue[cmd_queue_head].private;
	switch(i){
		case 0: {
			return t>>8;
		}
		case 1: {
			return t & 0xff;
		}
		default: return 'z';
	}
}

/** for .private=0xCDAB
 we print ABCD if len=4
 or AB if len=2
*/
uint8_t srl_direct_hex( uint8_t i ){
	uint16_t t = (uint16_t) cmd_queue[cmd_queue_head].private;
	switch(i){
		case 0: return hex_hi(t & 0xFF);
		case 1: return hex_lo(t & 0xFF);
        case 2: return hex_hi(t >> 8);
		case 3: return hex_lo(t >> 8);
		default: return 'z';
	}
}

/************************************
* generic serial logging function.
* uses private from the actual command as a pointer (array base).
* Unfortunately we need to know the string-length in advance
* For constant(-length) strings this is OK.
* We could set a (global or static variable is simplest) flag
* to notify the caller about the end of transmission.
* This way other comm functions could be left untouched.
************************************/
uint8_t srl_generic(uint8_t addr)
{
	return (cmd_queue[cmd_queue_head].private)[addr];
}

uint8_t srl_generic_atomic16_crc(uint8_t addr)
{
    // static uint8_t upper_byte; // replaced by ATOMIC_HELPER_SRL
    uint8_t t8;
    if (addr == 0) com.commcrc = 0xffff;

    if (addr >= DBG_LEN ){
        if (addr == DBG_LEN ) return com.commcrc & 0xff;//low byte
        return com.commcrc >> 8;//high byte
    }

    if (addr & 0x01){//odd memory address
        t8 = *ATOMIC_HELPER_SRL;
    } else { // even address
        *ATOMIC_HELPER_SRL = (cmd_queue[cmd_queue_head].private)[(addr + 1)];//save next byte to atomic read
        t8 = (cmd_queue[cmd_queue_head].private)[addr];
    }
    com.commcrc = _crc16_update( com.commcrc, t8);
	return t8;
}

/* same as above, but private points to PROGMEM (flash), eg. PSTR("...") */
uint8_t srl_generic_P(uint8_t addr)
{
	return PRG_RDB(&(cmd_queue[cmd_queue_head].private[addr]));
}

/* sends data direct from EEPROM in triggerlog format
*  240 byte data+ 1 bytetriggerlog type(16) + 2 byte crc + 1 byte HDLC_END
*  if EEPROM busy returns with CRC fail
*/
uint8_t srl_generic_E(uint8_t i)
{
    uint8_t v;
    if( i < (TRIGGERFRAME_READ_EEPROM_FIX + 3 ) ){
        if (i == ( TRIGGERFRAME_READ_EEPROM_FIX + 2 ) ) v = READ_EEPROM_STUFFED_TYPE_COMMAND-128;//answer type
        else if ( i == TRIGGERFRAME_READ_EEPROM_FIX + 1 ) v = (uint8_t)(uint16_t)cmd_queue[cmd_queue_head].private;//address low byte
        else if ( i == TRIGGERFRAME_READ_EEPROM_FIX ) v = (uint8_t)((uint16_t)cmd_queue[cmd_queue_head].private>>8);//address high byte
        else{
		if (eeprom.flags & _BV(active)) v = 0;//EEPROM busy send zeros. Some special value would make more sense...
		else v = eeprom_read_byte((uint8_t*)((uint16_t)(cmd_queue[cmd_queue_head].private)+i));
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( i >= ( TRIGGERFRAME_READ_EEPROM_FIX + 5 ) ){
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
        }else {
		if (eeprom.flags & _BV(active)) v = 0;//EEPROM busy send bad crc
		else v = (i == ( TRIGGERFRAME_READ_EEPROM_FIX + 3 ) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 0;
}

// helper funtion used to access flash above the 16 bit address limit
// TODO: check that it works
uint8_t srl_generic_P_upper(uint8_t addr)
{
	uint32_t v;
	v = 0x10000 | (uint16_t)&cmd_queue[cmd_queue_head].private[addr];
	return pgm_read_byte_far(v);
}

void change_page(uint8_t page) {
	if (page > MT_PAGE_IDX_LAST) {
		page = MT_PAGE_IDX_DUMMY;
	}
	debug.megatune_page = page;
}

/** ===============================================
 * make sure comm.h MT_PAGE_IDX_... and pages_size match.
 * Changing these breaks some tuningsoftwares (SIPR ??)
 */

#define MT_PAGE_PTR_DUMMY (uint8_t *)NULL

PGM_VOID_P const pages_ptr[] PROGMEM = {
 	(uint8_t *)&config
#ifdef HWMAP
 	,(uint8_t *)&hwmap.table
#else	//  sizeof(hwmap) bytes _must_ be transmitted to megatune
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef BOOSTCONTROL
	,(uint8_t *)&boost.tpstransfer_table
#else // no BOOSTCONTROL
 	, MT_PAGE_PTR_DUMMY
#endif
 	,(uint8_t *)&config_ext
#ifdef MAT_ENRICH
	,(uint8_t *)&ve.mat_enrich
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef A340E
 	,(uint8_t *)&at_config
#else
 	, MT_PAGE_PTR_DUMMY
#endif
	, ve.launch_rpmtable
    ,(uint8_t *)&config_s
    ,(uint8_t *)&ve.lambda
	,(uint8_t *)&ve.lambda // lambdaview !!!
 	,(uint8_t *)&ve.ve8_table
#ifdef CHARGE_TEMP
 	,(uint8_t *)&f_charge.f_charge
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef MAP_CORR
	,(uint8_t *)&ve.map_corr
#else
 	, MT_PAGE_PTR_DUMMY
#endif
	,(uint8_t *)&ve.mat_retard
	// following tables all share rpmkpa with ign_adv table:
	,(uint8_t *)&ign.adv_table
#ifdef BOOSTCONTROL
	,(uint8_t *)&boost.boost_ref_table
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef TOYOTA_1JZ_VVTI
 	,(uint8_t *)&vvti_map.vvti
#else
 	, MT_PAGE_PTR_DUMMY
#endif
	,(uint8_t *)&knock_table
};

PROGMEM const uint8_t pages_size[] = {
 	sizeof(struct config_t)
 	, HWMAP_BYTES
#ifdef BOOSTCONTROL
	,sizeof(boost.tpstransfer_table) + sizeof(boost.boostmaxtarget_table)
	+ sizeof(boost.boostpidrefpwm_table) + sizeof(boost.boostgear_rpmrange)
	+ sizeof(boost.anytrim1_array) + sizeof(boost.rotary_trailing_spark )
	+ sizeof(boost.inj_angle)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
	,sizeof(struct config_ext_t)
#ifdef MAT_ENRICH
	,224 //80 mat table + 2*64cam_table+16cam headers NOT sizeof(ve.mat_enrich) because PAGE_IDX_FIRST_RPMKPA
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef A340E
 	,sizeof(at_config)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef CLT_IGN_ADJUST
	,(sizeof(ve.launch_rpmtable) + 24) // +24 dirty hack to dump clt_ign_adjust + clt/kpa_range;
#else
	,sizeof(ve.launch_rpmtable)
#endif	
    ,sizeof(config_s)
	,sizeof(ve.lambda)
 	,sizeof(ve.lambda)
 	,sizeof(ve.ve8_table)
#ifdef CHARGE_TEMP
 	,sizeof(f_charge.f_charge)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef MAP_CORR
	,sizeof(ve.map_corr)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
	,sizeof(ve.mat_retard)
	// following tables all share rpmkpa with ign_adv table:
	,sizeof(ign.adv_table)
#ifdef BOOSTCONTROL
 	,sizeof(boost.boost_ref_table)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef TOYOTA_1JZ_VVTI
 	,sizeof(vvti_map.vvti)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
 	,sizeof(knock_table) // was: ,sizeof(knock_table.knock_noise)
};

uint8_t MEGATUNE_PAGE_SIZE(uint8_t page)
{
    return pgm_read_byte(pages_size + page);
}

uint8_t* MEGATUNE_PAGE_POINTER(uint8_t page)
{
    return (uint8_t*)pgm_read_word(pages_ptr + page);
}

/*
* calculates the 16 bit crc for megatune page with RPM/KPA header
*/
uint16_t megatune_page_crc(uint8_t idx)
{
    if (  idx > MT_PAGE_IDX_LAST) return 0xffff;//id not valid
    uint8_t cnt;
    uint8_t* page_ptr = MEGATUNE_PAGE_POINTER(idx);
    uint8_t page_len = MEGATUNE_PAGE_SIZE(idx);
    uint16_t crc = 0xffff;

    if ( page_ptr == MT_PAGE_PTR_DUMMY ) return 0xffff;
    for ( cnt=0 ; cnt<page_len ; cnt++){
#ifdef EXPENSIVE_SRAM_CRC
        crc = _crc16_update( crc, page_ptr[cnt]);
#else
        crc += page_ptr[cnt];
#endif
    }
    if (!(page_has_rpmkpa(idx))) return crc;

    struct bins_pointer_t bins_p;
	get_bins_pointer(idx, &bins_p);

    page_ptr = bins_p.p_rpm_range;
    page_len = bins_p.rpm_size;
    for ( cnt=0 ; cnt<page_len ; cnt++){
#ifdef EXPENSIVE_SRAM_CRC
        crc = _crc16_update( crc, page_ptr[cnt]);
#else
        crc += page_ptr[cnt];
#endif
    }
    page_ptr = bins_p.p_kpa_range;
    page_len = bins_p.kpa_size;
    for ( cnt=0 ; cnt<page_len ; cnt++){
#ifdef EXPENSIVE_SRAM_CRC
        crc = _crc16_update( crc, page_ptr[cnt]);
#else
        crc += page_ptr[cnt];
#endif
    }
    return crc;
}

/*
* returns the length for megatune page with RPM/KPA header
*/
uint8_t megatune_page_len(uint8_t idx)
{
    if (  idx > MT_PAGE_IDX_LAST) return 0;//id not valid
    uint8_t page_len = MEGATUNE_PAGE_SIZE(idx);
    if ( page_len == MT_PAGE_SIZE_DUMMY ) return 0;
    if (!(page_has_rpmkpa(idx))) return page_len;

    struct bins_pointer_t bins_p;
	get_bins_pointer(idx, &bins_p);
    page_len += bins_p.rpm_size;
    page_len += bins_p.kpa_size;
    return page_len;
}

/*
* sends the data from the private function with byte stuffing
* works only with parameterless private function which don use crc
*/
uint8_t sendStuffedSimple(uint8_t cnt)
{
    uint8_t size= cmd_queue[cmd_queue_head].len;
    uint8_t v;
    if (cnt <= (size-4)){
        if (cnt == (size-4)){
            v= (cmd_queue[cmd_queue_head].privw != 0) ? (uint8_t)cmd_queue[cmd_queue_head].privw : SIMPLE_MSG_STUFFED_TYPE_COMMAND-128;
        }else if (cnt == (size-5)){
            v= (cmd_queue[cmd_queue_head].privw != 0) ? (cmd_queue[cmd_queue_head].privw >>8 ) : 0;
        }else{
            v= (cmd_queue[cmd_queue_head].privatefunc)(cnt);
        }
	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    } else if(cnt >= size-1){
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
	} else { // i == 3 or 4
		v = (cnt == (size-3)) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian as in SIPR
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
	}

	return 's';
}

/*
* sends all page len and crc in triggerframe format
*/
uint8_t send_page_len_crc_stuffed (uint8_t cnt)
{
    uint8_t v;
    if( cnt <= MT_PAGE_CRC_PACKET_LEN-4 ){
        if (cnt == MT_PAGE_CRC_PACKET_LEN-4) v = GET_PAGE_LEN_CRC_STUFFED_TYPE_COMMAND-128;
        else if ( cnt < (2*(MT_PAGE_IDX_LAST+1))){//send all page crc
            if ( (cnt & 0x01) == 0){//even means calulate crc, store high byte and send low byte
                uint16_t crc = megatune_page_crc(cnt>>1);
                cmd_queue[cmd_queue_head].privch.a = crc >>8;
                v = (uint8_t)crc;
            }else{// just send the prestored crc low byte
                v = cmd_queue[cmd_queue_head].privch.a;
            }
        }else{ //send all page len
            v = megatune_page_len(cnt-(2*(MT_PAGE_IDX_LAST+1)));
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( cnt >= (MT_PAGE_CRC_PACKET_LEN-1) ){
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
        } else { // MT_PAGE_CRC_PACKET_LEN-1,-2
		v = (cnt == (MT_PAGE_CRC_PACKET_LEN-3) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 'C';
}
/**
*fill the command buffer (to send some bytes)
*/
// MODBUS_ON_COMM == 2

#ifdef MODBUS_SEND_BYTES_ATBOOT
void command_buffer(void)
{
    uint8_t idx;
    idx = com.uart1_tx_idx = 8;//buffer size

    //test command to start: 0x04 0x06 0x20 0x00 0x00 0x02 0x03 0x9E
    com.uart1_tx_buff[--idx] = 0x04;
    com.uart1_tx_buff[--idx] = 0x06;
    com.uart1_tx_buff[--idx] = 0x20;
    com.uart1_tx_buff[--idx] = 0x00;
    com.uart1_tx_buff[--idx] = 0x00;
    com.uart1_tx_buff[--idx] = 0x02;
    com.uart1_tx_buff[--idx] = 0x03;
    com.uart1_tx_buff[--idx] = 0x9E;
}
#endif

/***************************************************************************/
/*** comm                                                                ***/
/***************************************************************************/
void comm(uint8_t rx_data) {
  static uint8_t rx_offset = 0;
  static uint8_t rx_cmd = 0;

  switch (com.comm_state) {
  case OFFSET:
    rx_offset = rx_data;  // the offset for storing the new byte
    com.comm_state = VALUE;
    break;

  case VALUE:
    switch (rx_cmd) {
	  /* Separate command for each page write / read
	  * => No config/other overwrite on ecu reset. */
    case 'X': // forcing: channel + value
		debug.forced_channel = rx_offset;
		debug.forced_value = rx_data;
		break;
	  case 'O':
		change_page(MT_PAGE_IDX_CONFIG);
		store_page(rx_offset, rx_data);
		break;
	  case 'G':
		change_page(MT_PAGE_IDX_CONFIGEXT);
		store_page(rx_offset, rx_data);
		break;
	  case 131:
		change_page(MT_PAGE_IDX_CONFIG_SHARED);
		store_page(rx_offset, rx_data);
		break;
	  case 'H':
		change_page(MT_PAGE_IDX_HWMAP);
		store_page(rx_offset, rx_data);
		break;
	  case 'T':
	    change_page(MT_PAGE_IDX_BOOST);
		store_page(rx_offset, rx_data);
		break;
      case 'J':
		change_page(MT_PAGE_IDX_VE);
		store_page(rx_offset, rx_data);
		break;
      case 'N':
		change_page(MT_PAGE_IDX_IGNADV);
		store_page(rx_offset, rx_data);
		break;
      case 'L':
		change_page(MT_PAGE_IDX_LAMBDAVIEW);
		store_page(rx_offset, rx_data);
		break;
	case ')':
		change_page(MT_PAGE_IDX_LAMBDA);
		store_page(rx_offset, rx_data);
		break;
#ifdef A340E
	  case 'U':
		change_page(MT_PAGE_IDX_A340E);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef TOYOTA_1JZ_VVTI
	  case 'Z':
		change_page(MT_PAGE_IDX_1JZVVTI);
		store_page(rx_offset, rx_data);
		break;
#endif

	  case 'F':
		change_page(MT_PAGE_IDX_KNOCK);
		store_page(rx_offset, rx_data);
		break;

#ifdef BOOSTCONTROL
      case 'Y': // boost
		change_page(MT_PAGE_IDX_BOOST_REF);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef MAT_ENRICH
      case 'E': // temperatures
		change_page(MT_PAGE_IDX_TEMPERATURE);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef CHARGE_TEMP
      case 'E': // temperatures
		change_page(MT_PAGE_IDX_CHARGE);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef MAP_CORR
	case 'I': // map_corr
		change_page(MT_PAGE_IDX_MAPCORR);
		store_page(rx_offset, rx_data);
		break;
#endif

	case 'W': // mat_retard
		change_page(MT_PAGE_IDX_MATRET);
		store_page(rx_offset, rx_data);
		break;

	case ']':
		change_page(MT_PAGE_IDX_LAUNCHRPM);
		store_page(rx_offset, rx_data);
		break;

    case READ_SRAM_STUFFED_TYPE_COMMAND:{// sends page data in triggerlog format
            // rxdata stores the ID of the page if rx_offset ==0xff else ofset is the high address, rxdata is the low address
            uint8_t len;
            void* addr;
            if (rx_offset == 0xff){
                pushfuncpar(&SendPage_data_stuffed,0xFF,(void*)(0xff00|rx_data));//send the data(and rpmkpa) bytestuffed
                len = 0;
                addr = (void*)MEGATUNE_PAGE_POINTER(rx_data);
            }else{
                len = TRIGGERFRAME_READ_LEN_FIX;
                addr = (void*)( (rx_offset<<8) | rx_data);
            }
            pushfuncpar(&SendPage_stuffed,6+len,addr);// send the header
        }
        break;

      default:
        break;
    }
    com.comm_state = CMD; // back to normal mode
    break;

  default:
    switch (rx_data) {
    case 'A': // send all realtime values
	  aimdisable();
	  pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNT);
      break;
    case 'B': // store to eeprom
	save_to_storage();
      break;

    case 'C': // test communications
      pushfunc(&MTtestComm, 1);
      break;
    case 'D': // send all realtime values
	  aimdisable();
	  pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNTMORE+2);
      break;
/* do NOT reply in megatune mode because ECU reboot (=>megatune mode) confuses VT (will think still in triggerframe mode)
    case READ_REALTIME_STUFFED_TYPE_COMMAND: // 32+128=160 send all realtime values in triggerlog format
	  aimdisable();
	  pushfunc(MTsendRTvar_triggerlog, MT_RTVAR_COUNTMORE+4);
      break;
*/
    case 'Q': // send code version
	  aimdisable();
      pushfunc(MTcodeVer, MT_CODEVER_LEN);
      break;

    case 'S': // send textual signature
	  aimdisable();
      pushfunc(&sendSignature, 26);
      break;

    case 'v': // send page
	  pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);  // don't know the length yet
      break;

	/* Separate command for each page write / read
	* No config/other overwrite on ecu reset. */

	// config read
	case 'o':
		aimdisable();
		change_page(MT_PAGE_IDX_CONFIG);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// config_ext read
	case 'g':
		change_page(MT_PAGE_IDX_CONFIGEXT);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// config_s read
	case 130://'�'
		change_page(MT_PAGE_IDX_CONFIG_SHARED);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// h[] table read
	case 'h':
		change_page(MT_PAGE_IDX_HWMAP);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// tps transfer and boosttable read
	case 't':
		change_page(MT_PAGE_IDX_BOOST);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// ve ref table read
	case 'j':
		change_page(MT_PAGE_IDX_VE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// advance table read
	case 'n':
		change_page(MT_PAGE_IDX_IGNADV);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// lambda table read
	case 'l':
		change_page(MT_PAGE_IDX_LAMBDAVIEW);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
	case '(':
		change_page(MT_PAGE_IDX_LAMBDA);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#ifdef A340E
	case 'u':
		change_page(MT_PAGE_IDX_A340E);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef TOYOTA_1JZ_VVTI
	case 'z':
		change_page(MT_PAGE_IDX_1JZVVTI);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

	case 'f':
		change_page(MT_PAGE_IDX_KNOCK);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

    // boost Ref table read
#ifdef BOOSTCONTROL
	case 'y':
		change_page(MT_PAGE_IDX_BOOST_REF);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef MAT_ENRICH
	case 'e':
		change_page(MT_PAGE_IDX_TEMPERATURE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef CHARGE_TEMP
	case 'e':
		change_page(MT_PAGE_IDX_CHARGE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef MAP_CORR
	case 'i':
		change_page(MT_PAGE_IDX_MAPCORR);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

	case 'w':
		change_page(MT_PAGE_IDX_MATRET);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	case '[':
		change_page(MT_PAGE_IDX_LAUNCHRPM);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

    case 'X': // forcing: channel + value
	case 'G': // config_ext write
	case 131: // config shared write
	case 'H': // h[] table write
	case 'T': // tps transfer and boosttable write
	case 'J': // ve ref table write
	case 'N': // advance table write
	case 'L': // lambda table write
	case ')': // raw lambda table write
	case 'E': // tempt table write
	case 'U': // TODO: ifdef A340E
	case 'F': // TODO: ifdef KNOCK_ALTERNATIVE
	case 'Z': // TODO: ifdef TOYOTA_1JZ_VVTI
	case 'Y': // boost Ref table write TODO: ifdef BOOSTCONTROL
	case 'W':
	case 'I':
	case 'O': // config write
	case ']': // launchrpm(wheelspeed) table
    case READ_SRAM_STUFFED_TYPE_COMMAND: /* send the selected table in triggerlog format ( contains the address or pagename )*/
		com.comm_state = OFFSET;
		rx_cmd = rx_data;
		break;
    case GET_PAGE_LEN_CRC_STUFFED_TYPE_COMMAND: /* send the selected table in triggerlog format ( contains the address or pagename )*/
		pushfunc(&send_page_len_crc_stuffed,MT_PAGE_CRC_PACKET_LEN );
        break;

	default:
		break;
    }
  }
}

/* set needcrc to 2 if crc needed at the end of transfer
 * only makes sense for commands that send data to megatune, eg send_page()
 * (NOT for those that store)
 * Giant hack: we change the cmd_queue_head command to send commcrc after us
 */
static inline void terminate_transfer( uint8_t needcrc ) {
	cmd_queue[cmd_queue_head].count = 0xFF; // will become 0 in ringbuf_fill()
	cmd_queue[cmd_queue_head].len = needcrc ;
	cmd_queue[cmd_queue_head].func = (void*)0; // send 2 byte comm.crc
	cmd_queue[cmd_queue_head].private = (void*)(&com.commcrc);
}

/** send config and tables to vemstune */
uint8_t send_page(uint8_t i) {
	uint8_t val, size, page;

    uint16_t priv=(uint16_t)cmd_queue[cmd_queue_head].private;
	uint8_t needcrc=(priv>>8) ? 0 : 2; //set priv to zero if crc needed at the end of transfer
	page = (uint8_t)priv;
	size = MEGATUNE_PAGE_SIZE(page);

	if (page == MT_PAGE_IDX_DUMMY)
	    return(0);

	if( i < size ){
		uint8_t *tbl = MEGATUNE_PAGE_POINTER(page) ;
		val = tbl[i];

		if( page == MT_PAGE_IDX_LAMBDAVIEW ) {
		    return comm_getLambdaView(val);
		}

		if(page == MT_PAGE_IDX_BOOST) {
			// after t[] (with size TPSTRANSFERTABLE_SIZE) there comes
			// b[] (with size BOOSTCONTROL_GEARS * VE_SIZE_RPM):
			if( i >= TPSTRANSFERTABLE_SIZE ) {
				val = boost.boostmaxtarget_table[ i - TPSTRANSFERTABLE_SIZE ];
			}
		}

		if (!(page_has_rpmkpa(page)) && ( i >= size-1 )) {
			// don't need rpm/kpa headers?
			// so terminate
			terminate_transfer(needcrc);
		}

		return val;
	}

	val = 0; // suppress gcc warning

	struct bins_pointer_t bins_p;
	get_bins_pointer(page, &bins_p);

	i -= size;
	if (i < bins_p.rpm_size) {
		return bins_p.p_rpm_range[i];
	}
	i -= bins_p.rpm_size;
	if (i < bins_p.kpa_size )
		val = bins_p.p_kpa_range[i];
	if (i >= (bins_p.kpa_size - 1))
		terminate_transfer(needcrc);

	return val;
}

/**
 * some tables need special calculation
 */
void store_page(uint8_t addr, uint8_t value) {
    if ( com.status & _BV( MEGATUNE_WRITE_DISABLED) ) return;
	uint8_t size, page;

	page = debug.megatune_page;
	size = MEGATUNE_PAGE_SIZE(page);

	if (page == MT_PAGE_IDX_DUMMY)
        return;

	if (addr < size) {
		uint8_t *tbl;

		if(page == MT_PAGE_IDX_BOOST) {
			// after t[] there comes b[]:
			if( addr >= TPSTRANSFERTABLE_SIZE ){ // write to 2nd line
				boost.boostmaxtarget_table[ addr - TPSTRANSFERTABLE_SIZE ] = value;
				return;
			}
		}

		if(page == MT_PAGE_IDX_LAMBDAVIEW) {
		    value = comm_setLambdaView(value);
 		}

		// if( page == MT_PAGE_IDX_VE ) =>ve.ve8_table[addr] = value; //not special any longer
		tbl = MEGATUNE_PAGE_POINTER(page);
		tbl[addr] = value;
		return;
	}

	/*
	 * rpm_range and kpa_range arrays live at the end
	 */
	// need rpm/kpa headers?
	if(page_has_rpmkpa(page))
	{
		struct bins_pointer_t bins_p;
		get_bins_pointer(page, &bins_p);

		addr -= size;
		if ( addr < bins_p.rpm_size ) {
			bins_p.p_rpm_range[addr] = value;
			return;
		}
		addr -= bins_p.rpm_size;
		if (addr < bins_p.kpa_size )
			bins_p.p_kpa_range[addr] = value;
	}
}

/** only called for pages where page_has_rpmkpa() is true */
void get_bins_pointer(uint8_t page, struct bins_pointer_t *bins_pointer)
{
	switch (page)
	{
	case MT_PAGE_IDX_KNOCK: // should not be needed, since page_has_rpmkpa() is false as a special case
		bins_pointer->rpm_size = 0; // KNOCK_SIZE_RPM;
		bins_pointer->kpa_size = 0; // KNOCK_SIZE_KPA;
		bins_pointer->p_rpm_range = (uint8_t*)&knock_table; // .rpm_range;
		bins_pointer->p_kpa_range = (uint8_t*)&knock_table; // .kpa_range;
		break;
	case MT_PAGE_IDX_BOOST_REF: // detached from ignadv bins since 1.1.63 (2009-11-11)
		bins_pointer->rpm_size = IGNADV_SIZE_RPM;
		bins_pointer->kpa_size = IGNADV_SIZE_KPA;
		bins_pointer->p_rpm_range = boost.boost_ref_rpm_range;
		bins_pointer->p_kpa_range = boost.boost_ref_kpa_range;
		break;
	case MT_PAGE_IDX_IGNADV:
	case MT_PAGE_IDX_1JZVVTI:
		// ign adv bins
		bins_pointer->rpm_size = IGNADV_SIZE_RPM;
		bins_pointer->kpa_size = IGNADV_SIZE_KPA;
		bins_pointer->p_rpm_range = ign.adv_table_rpm_range;
		bins_pointer->p_kpa_range = ign.adv_table_kpa_range;
		break;
	// all 8x8 could use: p_rpm_range= page+64 and p_kpa_range = page+72
	case MT_PAGE_IDX_LAMBDA:
	case MT_PAGE_IDX_LAMBDAVIEW:
		// lambda bins
		bins_pointer->rpm_size = LAMBDA_SIZE_RPM;
		bins_pointer->kpa_size = LAMBDA_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.lambda_rpm_range;
		bins_pointer->p_kpa_range = ve.lambda_kpa_range;
        break;
#ifdef MAP_CORR
	case MT_PAGE_IDX_MAPCORR:
		// map corr bins
		bins_pointer->rpm_size = MAP_CORR_SIZE_RPM;
		bins_pointer->kpa_size = MAP_CORR_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.map_corr_rpm_range;
		bins_pointer->p_kpa_range = ve.map_corr_kpa_range;
		break;
#endif
	case MT_PAGE_IDX_VE:
	case MT_PAGE_IDX_CHARGE:
		// ve bins
		bins_pointer->rpm_size = VE_SIZE_RPM;
		bins_pointer->kpa_size = VE_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.rpm_range;
		bins_pointer->p_kpa_range = ve.kpa_range;
		break;
	case MT_PAGE_IDX_MATRET:
		// lambda bins
	default:// cheat to prevent gcc varning
		bins_pointer->rpm_size = LAMBDA_SIZE_RPM;
		bins_pointer->kpa_size = LAMBDA_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.mat_range;
		bins_pointer->p_kpa_range = ve.tps_range;
        break;
	}
}

/** because of byte stuffing, we push ourself. Min 2 bytes sent.
With some probability when i==1 we add 4 bytes: type(=1) + CRC + frame(=7E)
Called with 2 byte len automaticly close packet with 1/128 probability
*/
uint8_t TriggerlogPrimtrig( uint8_t i )
{
    com.status |= _BV(PRIMTRIG_NOT_CLOSED);//originally com.primtrig_not_closed = 1;
	uint8_t v;
	uint16_t timestamp = (uint16_t)(cmd_queue[cmd_queue_head].private);

	if( i < 3 ){ // 0, 1 or 2
		if(i==0) v = timestamp >> 8; // send upper byte
		else if(i==1){
			v = timestamp & 0xFF; // send lower byte
			// with 1/128 probability extend length to 6 bytes:
			if( (v & 0x1F) == 0 ) cmd_queue[cmd_queue_head].len = 6; // earlier (v & 0x7F) primtrig took too long to show up without sectrig, appr 128 pulses
		} else v = 1; // i ==2, type = 1: means CRC only
		ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else { // 3 and 4 are CRC
		if( i == 5 ){
			ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
            com.status &= ~_BV(PRIMTRIG_NOT_CLOSED);//originally com.primtrig_not_closed = 0;
		} else { // 3 or 4
			v = (i == 3) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
			ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
	}
	return 'P'; // cheaper than return i, same as return 0
}
/** more than 6 bytes might be sent because of byte stuffing: we push ourself */
uint8_t TriggerlogAnytrig( uint8_t i )
{
	uint8_t v;
	uint16_t timestamp = (uint16_t)(cmd_queue[cmd_queue_head].private);
	if(i <= 2){
	  switch(i){
		case 0: v = timestamp >> 8; break; // send upper byte (Big endian)
		case 1: v = timestamp & 0xFF; break; // send lower byte
		default: v=cmd_queue[cmd_queue_head].privch.a; break; // type is from privch.a
	  }
	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else if(i>=5){
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
        com.status &= ~_BV(PRIMTRIG_NOT_CLOSED);//originally com.primtrig_not_closed = 0;
	} else { // i == 3 or 4
		v = (i == 3) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian as in SIPR
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
	}
	return 'T';
}

/** Apparently unused
/// just like TriggerlogAnytrig, but have 1 byte extra parameter and can't use after unfinished ptrigs
uint8_t TriggerlogAnytrigPar( uint8_t i )
{
	uint8_t v;
	uint16_t timestamp = (uint16_t)(cmd_queue[cmd_queue_head].private);
	uint8_t pos = cmd_queue[cmd_queue_head].len - i;

	if(pos >= 4){
	  switch(pos){
		case 7: v = cmd_queue[cmd_queue_head].privch.b; break; // extra parameter upper byte (Big endian)
		case 6: v = timestamp >> 8; break; // send upper byte (Big endian)
		case 5: v = timestamp & 0xFF; break; // send lower byte
		default: v=cmd_queue[cmd_queue_head].privch.a; break; // type is from privch.a
	  }
	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else if(pos<=1){
		ringbuf_push_HDLC_END_commcrc_ffff(); // send framing char, do NOT stuff
        com.status &= ~_BV(PRIMTRIG_NOT_CLOSED);//originally com.primtrig_not_closed = 0;
	} else { // pos == 3 or 2
		v = (pos == 3) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian as in SIPR
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
	}
	return 'T';
}
*/

/***************************************************************************/
/*** pushfunc                                                            ***/
/***************************************************************************/
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len) {
	uint8_t tail, next_tail;
	tail = cmd_queue_tail;
	next_tail = (tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

  /* len starts at 0 */
//	if (len > 0) // this looks like a hack - it is! IT WAS!
//		len = len - 1;
  cmd_queue_tail = next_tail; // update this (global) var first: so if a trigger or wheelspeed irq comes here and calls pushfuncpar2(), less likely to get mixed parameters in same bucket. When cmd_queue_tail is read from userspace, it will see all other data filled in:
  cmd_queue[tail].func = f;
  cmd_queue[tail].len = len;
  cmd_queue[tail].count = 0;
  cmd_queue[tail].private = 0;
//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}

/** Currently void, since cmd_queue overrun cannot be handled by caller. Caller could check BEFORE calling.
** same as pushfunc(), but more parameters
*/
void pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint8_t p2)
{
	uint8_t tail, next_tail;
	tail = cmd_queue_tail;
	next_tail = (tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return; // return -1; // whoops, queue is full

  cmd_queue_tail = next_tail; // update this (global) var first: so if a trigger or wheelspeed irq comes here and calls pushfuncpar2(), less likely to get mixed parameters in same bucket
  cmd_queue[tail].func = f;
  cmd_queue[tail].len = len;
  cmd_queue[tail].count = 0;
  cmd_queue[tail].private = par;
  cmd_queue[tail].privch.a = p2;

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return; // return 0;
}
//eg.: pushfuncpar2W(&srl_write_4byte_plus_end,7,(void*)((WRITE_EEPROM_STUFFED_TYPE_COMMAND - 128 ) | (eeprom.length <<8 )),(uint16_t)eeprom.p_eeprom);//send acknowledge on write (written cnt,address)
void pushfuncpar2W(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint16_t p2)
{
	uint8_t tail, next_tail;
	tail = cmd_queue_tail;
	next_tail = (tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return; // return -1; // whoops, queue is full

  cmd_queue_tail = next_tail; // update this (global) var first
  cmd_queue[tail].func = f;
  cmd_queue[tail].len = len;
  cmd_queue[tail].count = 0;
  cmd_queue[tail].private = par;
  cmd_queue[tail].privw = p2;
//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return; // return 0;
}

// compatibility function with less parameters
void pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, void * par)
{
	pushfuncpar2(f, len, par, 0);
}
/**
* Dumps a few tables, starting with wherever p and p+1 points to (travelling forward).
* makes sure entries < CMD_QUEUE_SIZE (not equal !!!)
*/
void pushfuncTableDump(const uint8_t *p, uint8_t entries)
{
	// uint8_t allowed = CMD_QUEUE_SIZE-1; // not more than this many, please. Not needed if entries n< CMD_QUEUE_SIZE
	do {
		// pgm_read_word_near( p ); // endianess ?
		uint8_t table = pgm_read_byte_near( p++ );
		uint8_t lines = pgm_read_byte_near( p++ );
		pushfuncpar2(&TableDump, 0xFF, (void*)( (table << 8) + lines ), 0 );
	} while (--entries);
}
/* "small" wrapper because otherwise menu.c does not know size... Avoided because it bloated FLASH SIZE by +500 bytes !
void pushfuncTableDump1( void )
{
	 pushfuncTableDump( TABLES_TO_DUMP1, sizeof(TABLES_TO_DUMP1) >> 1 );
} */

/***************************************************************************/
/*** initUART                                                            ***/
/***************************************************************************/
void initUART(void) {
  /* reset the baudrate flag. 0-d by gcc at startup, but ... */
  new_baudrate=0;

  com.comm_state = CMD;

  /* rx enable, tx enable */
  //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
  UCSR0B = _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
  /* 8 databit, 1 stopbit */
  UCSR0C = _BV(UCSZ01) | _BV(UCSZ00)
#if MODBUS_ON_COMM == 1
   | (COMM_8E1 ? _BV( UPM01 ) : 0)
#endif
   ; /* | _BV( UPM01 ) parity enabled: even parity if UPM00 not set */

  /* 19200 baud @16MHz : 1000000 / (51+1) = 19231.
  * Set baud AFTER UCSR0B and UCSR0C because of shared registers on some atmegas - although theoretically NOT in atmega128 */
  UBRR0H = 0;
  UBRR0L = 0x33;

  cmd_queue_tail = cmd_queue_head = 0;
	com.commcrc = 0xffff; // 12 bytes flash. Not needed for D command (set at offset 0), but useful for mde40 logging and possibly others in future
    com.status =0; //originally com.primtrig_not_closed = 0;
}
/**
* initUART1 (1-wire)
* For 19200 and lower (9600) baud, polling from mainloop is (surprisingly) much more efficient than interrupts (which have prohibitive push/poll overhead).
* with the cheap if( UCSR1A & (1<<UDRE1) ) ... condition even the config.whatever need not be checked, only in init: if UART is initialized in init (TXEN1),
* (and not filled from interrupt because UDRIE1 interrupt not enabled) => the sending register will become empty, so just send next byte
*/
void initUART1(void)
{
#ifdef ONE_WIRE
    if ( IBUTTON_ENABLED ){// ibutton configured
     // rx enable
      //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
      UCSR1B = _BV(RXCIE1) | _BV(RXEN1) | _BV(TXEN1); // tx enable is needed
      /* 8 databit, 1 stopbit */
      UCSR1C = _BV(UCSZ11) | _BV(UCSZ10);
      UBRR1H = 0;      /* 9600 baud @16MHz */
      UBRR1L = 103;
      return;//GPSlog cannot work with 1wire
   }
#endif //ONE_WIRE
    uint8_t comm2_baud = gps_baud(); // "GPS baud" could be renamed to "second serial baudrate", used always except 1-wire/IBUTTON
    UBRR1H = 0;
    UBRR1L = comm2_baud; // does not hurt, even if UART1 is not enabled later
    // AIM standard says 19200,8n1 http://www.vems.hu/files/MembersPage/NanassyPeter/AIM_support/AIM-ECU%20protocol.pdf
    // also compatible with VEMS round (or displayonly round), but we leave it configurable now

#ifdef GPSLOG
    if ( (sd_bytestuffed() )//GPS_LOG works only with bytestuffed SD log
        && ( (config.gps_log & _BV(GPS_LOG_ON_COM1_ENABLE_BIT)) == 0 )//GPS_LOG on first serial disabled => enabled on second
    ){
        // rx enable
        //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
        UCSR1B = _BV(RXCIE1) | _BV(RXEN1) | _BV(TXEN1); // tx enable is needed ??
        /* 8 databit, 1 stopbit */
        UCSR1C = _BV(UCSZ11) | _BV(UCSZ10);
        com.gps.state = GPS_DATA_EMPTY;
        return; // initialized to receive NMEA GPS data
    }
#endif //GPSLOG
//init UART1
    if( COMM_8E1_or_8N1 && (( config.lcd_c0 & _BV(lcd_rs232aim))==0) ){ // AIM not on UART0 => UART1
        UCSR1B = _BV(TXEN1)
#if MODBUS_ON_COMM == 2         
        |  _BV(RXCIE1) | _BV(RXEN1) | _BV(UDRIE1) // enable RX via ISR, and TX empty interrupt (if NOT enabled, ml_highest_run() will notice UCSR1A & (1<<UDRE1) )
#endif //MODBUS_ON_COMM == 2
        ; // tx enable. If UDR1 = missing, then immediately sends an offending 0 byte
        /* 8 databit, 1 stopbit */
        UCSR1C = _BV(UCSZ11) | _BV(UCSZ10) | (COMM_8E1 ? _BV( UPM11 ) : 0); /* | _BV( UPM11 ) parity enabled: even parity as UPM10 not set */;
        return;
    }
}

/***************************************************************************/
/*** INTERRUPT: UART rx complete ***/
/***************************************************************************/
SIGNAL(USART0_RX_vect) { // formerly SIG_UART0_RECV
    uint8_t readstatus, received;
	readstatus = UCSR0A; // datasheet p183: UCSRA must be read before the receive buffer (UDR)
	received = UDR0; // UDR must be read in any case
  if(( readstatus & (_BV(FE0) | _BV(DOR0)) ) ==0) { // no frame or overrun errors
    // valid character received
    ringbuf_push(&comm_recbuf, received);
#if MODBUS_ON_COMM == 1
    if( menu_rs232.state == MENU_MODBUS ){ // timing is interesting in modbus RTU
            uint16_t tdiff, wallclock_last = com.incoming_CRC;
            com.incoming_CRC = tdiff = ( (etime.now_upper << 8) + (etime.now>>8)); // 1 msec
            tdiff -= wallclock_last; //  time diff since last char received
            if( tdiff >= MODBUS_TIMEOUT){  // finally 3 (4..6 msec) or more
                com.modbus_rx_cnt = com.modbus_rx_state = 0; // reset to framestart. What if not yet processed by userspace ?
                if(MODBUS_DEBUG_SIGNALTIMEOUT) pushfuncpar(&srl_generic_P, 1, PSTR("T"));
            }
    }
#endif
  }
}

#ifdef ONE_WIRE
void one_wire_rx(uint8_t readchar)
{
    uint8_t rx_cnt;
  	rx_cnt = one_wire.rx_cnt;
  	rx_cnt++;
  	if ((rx_cnt == 2) && (readchar == 0x33)) one_wire.rx_crc=0;
  	else{
  	    if (rx_cnt == 10){//crc incoming
  	        if (one_wire.rx_crc == readchar) one_wire.status = ONE_WIRE_CRC_OK;
  	        else one_wire.status = ONE_WIRE_CRC_FAIL;
  	    }else{
            if (rx_cnt == 4)
                one_wire.msB=readchar;
            else if (rx_cnt == 5)
                one_wire.ssB=readchar;
            one_wire.rx_crc = _crc_ibutton_update(one_wire.rx_crc,readchar);
  	    }
  	}
  	one_wire.rx_cnt = rx_cnt;

}
#endif //ONE_WIRE


/**
* INTERRUPT: UART1 (1-wire) rx complete or second GPS input
*/
SIGNAL(USART1_RX_vect ) { // formerly SIG_UART1_RECV

	uint8_t readchar;
	readchar = UDR1;	// read uart
	// check status register for frame or overrun errors
	if ( UCSR1A & (_BV(FE1) | _BV(DOR1)) ) {
		 // just and throw away the result
    		return;
	}
	if( com.comm1_relay_allowed ){ // mxw  sends chars received from 1wire to main uart0 (RS232)
	// inject into char-buf of UART0 directly
	// BEWARE: possibly into the middle of some bigger block, so only useful for testing
		ringbuf_push(&comm_sndbuf, readchar);
		UCSR0B |= _BV(UDRIE0); // enable transmit interrupt for UART0, so comm_sndbuf does not get stuck
  	}
#ifdef ONE_WIRE
    if ( IBUTTON_ENABLED ){// ibutton configured
	    one_wire_rx(readchar);
	    return;//GPSlog cannot work with 1wire
	}
#endif //ONE_WIRE
#ifdef GPSLOG
    if (sd_bytestuffed()){//GPS_LOG works only with bytestuffed SD log
        gps_parse(readchar); // process char from USART1
        return;
    }
#endif
#if MODBUS_ON_COMM == 2
    if(COMM_8E1_or_8N1){ // TODO: receive analog data (and wheelspeeds) from attiny828 
        serial_feed(readchar); // incoming modbus frame processing
        // alternatively some buffer like ringbuf_push(&comm_recbuf, readchar); but prolly not needed
    }
#endif //MODBUS_ON_COMM == 2
}

#if MODBUS_ON_COMM == 2
/**
* INTERRUPT: UART1 tx empty
* #if MODBUS_ON_COMM == 2
* 1. when tx_buf_idx is zero we fill the buffer with the command
* 2. send the command byte by byte by decrement tx_buf_idx
* 3. after then tx_buf_idx is 1 it will be 0xFF (to send only once)
*/
void uart1_send_from_buf(void)
{
    uint8_t tx_buf_idx = com.uart1_tx_idx; // number of bytes in com.uart1_tx_buff
 	if (tx_buf_idx == 0) { // finished. Cannot underflow from 00 to FF so >= 0xFF or 0xFE or similar check should not be needed
        UCSR1B &=~_BV(UDRIE1);// disable data register empty interrupt on usart1
        // UCSR1A = _BV(UDRE1); // clear interrupt flag, invalid for this
        // tx_buf_idx = 0xFF; let it remain 0
        return;
    }
    tx_buf_idx --; //first need to decrement when zero we don't need to send
    UDR1 = tx_buf_idx < UART1_BUFF_LEN ? com.uart1_tx_buff[tx_buf_idx] : tx_buf_idx ;// send preamble filler bytes if more bytes than buffer length
    com.uart1_tx_idx = tx_buf_idx ;
}
#endif //MODBUS_ON_COMM == 2

#if defined ONE_WIRE || MODBUS_ON_COMM == 2
/**
* INTERRUPT: UART1 (1-wire) tx empty
* Important: NOT enabled for AIM on UART1 !
*/
SIGNAL(USART1_UDRE_vect) { // formerly SIG_UART1_DATA
#if MODBUS_ON_COMM == 2
    if(COMM_8E1_or_8N1){
        uart1_send_from_buf();
        return;
    }
#endif
#ifdef ONE_WIRE
    uint8_t tx_cnt = one_wire.tx_cnt;
    UDR1 = pgm_read_byte( &(oneWireInit[tx_cnt++] ));//send the next byte
    if (tx_cnt == oneWireInitLen){//one_wire init finished
        UCSR1B &=~_BV(UDRIE1);// disable data register empty interrupt on usart1
        tx_cnt = 0;//reenable 1wire init
    }
    one_wire.tx_cnt = tx_cnt;
#endif
}
#endif //ONE_WIRE

/***************************************************************************/
/*** INTERRUPT: UART0 tx empty ***/
/***************************************************************************/
SIGNAL(USART0_UDRE_vect ) { // formerly SIG_UART0_DATA

	uint16_t d;

	d = ringbuf_shift(&comm_sndbuf);

	if (d>>8) { // buffer empty
                // this is tricky. To save on useless processing of the SIG_UART0_TRANS
                // interrupt we will enable that interrupt to signal us when this last byte
                // has been sent if and only if there is a baudrate change pending
		UCSR0B &= ~_BV(UDRIE0);    // disable TX interrupts
                if (new_baudrate!=0)
		  UCSR0B |= _BV(TXCIE0);    // enable TX complete interrupt
	} else {
		UDR0 = d;
	}

}

/***************************************************************************/
/*** INTERRUPT: UART0 tx complete ***/
/***************************************************************************/
SIGNAL(USART0_TX_vect) { // formerly SIG_UART0_TRANS

  // we can't change the baudrate when we get a TX empty interrupt because
  // the shift register still contains the last byte to be sent. So here we
  // check to make sure the TX interrupts are disabled (nothing buffered)
  // and if the new_baudrate flag is set then we're good to change the baudrate
  // furthermore since we don't want to waste any CPU cycles we will disable
  // our interrupt.

  // if the TX interrupt is enabled then someone must have sent MORE data while
  // we were waiting for the last byte to send and refilled the buffer so
  // disable our interrupt and bail
  if (UCSR0B & _BV(UDRIE0)) { // someone sent MORE data... Note that we might miss the TX-complete (and forget baudrate change) if unlucky
    UCSR0B &= ~_BV(TXCIE0);    // disable the TX complete interrupt
  }
  else {
    // if there is a baudrate change pending then make sure the cmd_queue is empty
    // and change the baudrate. Otherwise disable TX interrupts
    if ( (new_baudrate!=0) && (cmd_queue_tail==cmd_queue_head)) {
      change_baudrate(new_baudrate);
      new_baudrate=0;
      UCSR0B &= ~_BV(TXCIE0);    // disable the TX complete interrupt
#if MODBUS_ON_COMM == 1
      if( menu_rs232.state != MENU_MODBUS ) UCSR0C = _BV(UCSZ01) | _BV(UCSZ00); /* no | _BV( UPM01 ) => NO parity. Useful when leaving 8E1 mode*/
#endif
    }
  }
}

// Use for debugging
void put_char(uint8_t byte)
{
    while( !(UCSR0A & 0x20) )
        ;
    UDR0 = byte;
}

uint8_t comm_getLambdaView(uint8_t val)
{
    uint16_t mt_val;
    mt_val = mult16_8((val + LAMBDA_INT_OFFSET), 72) >> 7;
    mt_val = (mult16_8(div_65535_x(mt_val), 72) >> 7) - MT_LAMBDA_OFFSET;
    return mt_val & 0xFF;
}

uint8_t comm_setLambdaView(uint8_t val)
{
    uint16_t int_val;
    int_val = mult16_8((val + MT_LAMBDA_OFFSET), 72) >> 7;
    int_val = mult16_8(div_65535_x(int_val) , 72) >> 7;
    val = (int_val - LAMBDA_INT_OFFSET) & 0xFF;
    return val;
}
